@@ -2,6 +2,8 @@
 
 import ast
 import codeop
+from io import StringIO
+from tokenize import generate_tokens, STRING, TokenError
 
 CORO_NAME = "__corofn"
 CORO_DEF = f"async def {CORO_NAME}(): "
@@ -28,8 +30,38 @@ def exec_result(obj, local, stream):
         print(repr(obj), file=stream)
 
 
+class ReturnChecker(ast.NodeVisitor):
+    def __init__(self, filename):
+        super().__init__()
+        self.filename = filename
+
+    def visit_FunctionDef(self, node: ast.FunctionDef):
+        return  # skip functions
+
+    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):
+        return  # async functions too
+
+    def visit_Return(self, node: ast.Return):
+        raise SyntaxError(
+            "'return' outside function",
+            (self.filename, node.lineno, node.col_offset + 1, None),
+        )
+
+    def visit_Yield(self, node: ast.Yield):
+        raise SyntaxError(
+            "'yield' outside function",
+            (self.filename, node.lineno, node.col_offset + 1, None),
+        )
+
+    def visit_YieldFrom(self, node: ast.YieldFrom):
+        self.visit_Yield(node)  # handle in the same way as regular yield
+
+
 def make_tree(statement, filename="<aexec>", symbol="single", local={}):
     """Helper for *aexec*."""
+    # Check for returns and yields
+    ReturnChecker(filename).visit(statement)
+
     # Create tree
     tree = ast.parse(CORO_CODE, filename, symbol)
     # Check expression statement
@@ -50,6 +82,16 @@ def make_coroutine_from_tree(tree, filename="<aexec>", symbol="single", local={}
     return dct[CORO_NAME](**local)
 
 
+def get_non_indented_lines(source):
+    try:
+        for token in generate_tokens(StringIO(source).readline):
+            if token.type == STRING:
+                # .start and .end line numbers are one-indexed
+                yield from range(token.start[0], token.end[0])
+    except TokenError:
+        pass
+
+
 def compile_for_aexec(
     source, filename="<aexec>", mode="single", dont_imply_dedent=False, local={}
 ):
@@ -59,17 +101,17 @@ def compile_for_aexec(
         flags |= codeop.PyCF_DONT_IMPLY_DEDENT
 
     # Avoid a syntax error by wrapping code with `async def`
-    indented = "\n".join(line and " " * 4 + line for line in source.split("\n"))
+    # Disabling indentation inside multiline strings
+    non_indented = set(  # sets are faster for `in` operation
+        get_non_indented_lines(source)
+    )
+    indented = "\n".join(
+        (" " * 4 if i not in non_indented and line else "") + line
+        for i, line in enumerate(source.split("\n"))
+    )
     coroutine = CORO_DEF + "\n" + indented + "\n"
     interactive = compile(coroutine, filename, mode, flags).body[0]
 
-    # Check EOF errors
-    try:
-        compile(source, filename, mode, flags)
-    except SyntaxError as exc:
-        if exc.msg == "unexpected EOF while parsing":
-            raise
-
     return [make_tree(statement, filename, mode) for statement in interactive.body]
 
 
